/****************************************************************************
* 
* UARATIO.c - Routinen zur Ein- und Ausgabe auf der seriellen Schnittstelle
*
****************************************************************************/

#include <Arduino.h>
#include "EBusAna.h"
#include "Global.h"
#ifdef SD_LOGGING
#include <SPI.h>
#include <SD.h>
#endif /* #ifdef SD_LOGGING */

/****************************************************************************
* Definitionen
****************************************************************************/

#define OUTBUFLEN 40
#define OUT_SERIAL 1
#define OUT_SD 2

/****************************************************************************
* globale Daten
****************************************************************************/

static const char StatusErrTxt[] PROGMEM = "???";
static const char StatusTxt1[]   PROGMEM = "off";
static const char StatusTxt2[]   PROGMEM = "keine Aktion";
static const char StatusTxt5[]   PROGMEM = "Regler stop";
static const char StatusTxt6[]   PROGMEM = "Brauchwasserbereitung";
static const char StatusTxt7[]   PROGMEM = "Brauchwasserbereitung, Regler stop";
static const char StatusTxt11[]  PROGMEM = "Heizbetrieb";
static const char StatusTxt12[]  PROGMEM = "Heizen, Warmwasser";
static const char StatusTxt13[]  PROGMEM = "Emisionskontrolle";
static const char StatusTxt14[]  PROGMEM = "TUEV Funktion";
static const char StatusTxt15[]  PROGMEM = "Regler Stop";
static const char* const StatusText[] PROGMEM = {
  StatusTxt1,       /* 0x00 */
  StatusTxt2,       /* 0x01 */
  StatusErrTxt,     /* 0x22 */
  StatusErrTxt,     /* 0x33 */
  StatusTxt5,       /* 0x44 */
  StatusTxt6,       /* 0x55 */
  StatusTxt7,       /* 0x66 */
  StatusErrTxt,     /* 0x77 */
  StatusErrTxt,     /* 0x88 */
  StatusErrTxt,     /* 0x99 */
  StatusTxt11,      /* 0xAA */
  StatusTxt12,      /* 0xBB */
  StatusTxt13,      /* 0xCC */
  StatusTxt14,      /* 0xDD */
  StatusTxt15,      /* 0xEE */
  StatusErrTxt };   /* 0xFF */

static char OutBuf[ OUTBUFLEN ];
char Dezimalpunkt = ',';
static char OutPut = OUT_SERIAL;
#ifdef SD_LOGGING
File dataFile;
extern File logFile;
#endif /* #ifdef SD_LOGGING */

/*--------------------------------------------------------------------------*
* FlexPrint - Ausgabe von Puffer auf Serial oder SD
* Parameter: pBuf - Zeiger auf Puffer
* Ergebnis:  keines
*--------------------------------------------------------------------------*/

void FlexPrint( char *pBuf )

  {
#ifdef SD_LOGGING
  if( OutPut & OUT_SERIAL )
    {
#endif /* #ifdef SD_LOGGING */
//    if( *pBuf == '\n' )  
//      COMMAND_PORT.write( '\r' );
    COMMAND_PORT.write( pBuf  );
#ifdef SD_LOGGING
#ifdef SD_MES_LOG
    if( logFile )
      logFile.write( pBuf );
#endif /* #ifdef SD_MES_LOG */
    }
  if( (OutPut & OUT_SD) && dataFile )
    dataFile.write( pBuf );
#endif /* #ifdef SD_LOGGING */
  }


/*--------------------------------------------------------------------------*
* PrintNewLine - Zeilenumbruch ausgeben
* Parameter: keine
* Ergebnis:  keines
*--------------------------------------------------------------------------*/

void PrintNewLine( void )

  {
  FlexPrint( (char *)"\n");  
  }


/*--------------------------------------------------------------------------*
* UWtoASCII - Unsigned Wert in ASCII wandeln
* Parameter: Data - Wert
*            len  - Minimale Zahl der Stellen
* Ergebnis:  pStr - Zeiger auf Ergebnis
*--------------------------------------------------------------------------*/

static char *UWtoASCII( char *pBuf, uint16_t Data, uint8_t len )

  {
  do
    {
    *(--pBuf) = Data % 10 + '0'; 
    Data /= 10;
    if( len > 0 )
      len--;
    }
  while( Data != 0 );

  while( len-- > 0 )
    {
#ifdef DEBUG
    if( (pBuf <= &OutBuf[0]) || (pBuf > &OutBuf[OUTBUFLEN]) )
      COMMAND_PORT.println( F("Error in UWtoASCII") );
#endif /* #ifdef DEBUG */
    *(--pBuf) = '0'; 
    }
    
  return( pBuf );
  }


/*--------------------------------------------------------------------------*
* UWtoFloat - Unsigned Wert in Kommazahl (ASCII) wandeln
* Parameter: Data - Wert
*            Div  - Teiler für Nachkommastelle 
*            len  - Zahl der Nachkommastellen
* Ergebnis:  pStr - Zeiger auf Ergebnis
*--------------------------------------------------------------------------*/

static char *UWtoFloat( uint16_t Data, uint16_t Div, uint8_t len )

  {
  char *pBuf;
  uint16_t NachK, VorK;
  uint8_t i;

  pBuf = &OutBuf[ OUTBUFLEN - len - 1];
  VorK  = Data / Div;

  if( len > 0 )
    {
    NachK = Data % Div;
    NachK *= 2560/Div;

    for( i = 0; i < len; i++ )
      {
      *(pBuf++) = highByte( NachK ) + '0'; 
      NachK &= 0xFF;
      NachK *= 10;
      }

    *(pBuf++) = 0; 
    pBuf = &OutBuf[ OUTBUFLEN - len - 1];
    *(--pBuf) = Dezimalpunkt;
    }
  else
    *(--pBuf) = 0;

  pBuf = UWtoASCII( pBuf, VorK, 1 );

  return( pBuf );
  }


/*--------------------------------------------------------------------------*
* toHexByte - Wert in HExadezimal (ASCII) wandeln
* Parameter: Data - Wert
*            len  - Zahl der Zeichen
* Ergebnis:  pStr - Zeiger auf Ergebnis
*--------------------------------------------------------------------------*/

static char *toHexByte( uint16_t Data, uint8_t len )

  {
  char *pBuf = OutBuf + OUTBUFLEN;
  char ch;

  *(--pBuf) = 0;
  if( len & 1 )
    {
    *(--pBuf) = ' ';
    len--;
    }
    
  while( len-- > 0 )
    {
    ch = (Data & 0x000F) + '0';
    if( ch > '9' )
      ch += 'A' - '0' - 10;
    *(--pBuf) = ch;
    Data >>= 4;
    }

  return( pBuf );
  }


/*--------------------------------------------------------------------------*
* PrintFloat - Unsigned Wert als Kommazahl ausgeben
* Parameter: Data - Wert
*            Div  - Teiler für Nachkommastelle 
*            len  - Zahl der Nachkommastellen
* Ergebnis:  keines
*--------------------------------------------------------------------------*/

void PrintFloat( uint16_t Data, uint16_t Div, uint8_t len )
  {
  char *pBuf;

  pBuf = UWtoFloat( Data, Div, len );
  FlexPrint( pBuf );  
  }


/*--------------------------------------------------------------------------*
* PrintSFloat - Signed Wert als Kommazahl ausgeben
* Parameter: Data - Wert
*            Div  - Teiler für Nachkommastelle 
*            len  - Zahl der Nachkommastellen
* Ergebnis:  keines
*--------------------------------------------------------------------------*/

void PrintSFloat( int16_t Data, uint16_t Div, uint8_t len )
  {
  char *pBuf;

  if( Data < 0 )
    {
    pBuf = UWtoFloat( -Data, Div, len );
    *(--pBuf) = '-';
    }
  else
    pBuf = UWtoFloat( Data, Div, len );

  FlexPrint( pBuf );  
  }


/*--------------------------------------------------------------------------*
* PrintDec - Unsigned Wert Dezimal ausgeben
* Parameter: Data - Wert
* Ergebnis:  keines
*--------------------------------------------------------------------------*/

void PrintDec( uint16_t Data )
  {
  char *pBuf = OutBuf + OUTBUFLEN;

  *(--pBuf) = 0;
  
  pBuf = UWtoASCII( pBuf, Data, 1 );
  FlexPrint( pBuf );  
  }


/*--------------------------------------------------------------------------*
* PrintSDec - Signed Wert Dezimal ausgeben
* Parameter: Data - Wert
* Ergebnis:  keines
*--------------------------------------------------------------------------*/

void PrintSDec( int16_t Data )
  {
  char *pBuf = OutBuf + OUTBUFLEN;

  *(--pBuf) = 0;

  if( Data < 0 )
    {
    pBuf = UWtoASCII( pBuf, -Data, 1 );
    *(--pBuf) = '-';
    }
  else
    pBuf = UWtoASCII( pBuf, Data, 1 );

  FlexPrint( pBuf );  
  }


/*--------------------------------------------------------------------------*
* PrintHex - Git einen Wert Hexadezimal aus
* Parameter: Data - Wert
*            len  - Zahl der Zeichen, ungeraden -> Leerzeichen anhängen
* Ergebnis:  keines
*--------------------------------------------------------------------------*/
void PrintHex( uint16_t Data, uint8_t len )

  {
  char *pBuf;

  pBuf = toHexByte( Data, len );
  FlexPrint( pBuf );  
  }


/*--------------------------------------------------------------------------*
* PrintData - Gibt einem Messwert aus
* Parameter: Data - Zeiger auf die Daten
* 			     Type - T�p der Daten
* Ergebnis:  keines
*--------------------------------------------------------------------------*/

void PrintData( union DATA_TYPE Data, uint8_t Type )

  {
  static const char *Wochentag[] = { "Montag", "Dienstag", "Mittwoch", "Donnerstag", "Freitag", "Samstag", "Sonntag" };
  char *pBuf;

  switch( Type )
	  {
	  case _BYTE:  PrintSDec(   Data.B  ); break;
	  case _UBYTE: PrintDec(    Data.UB ); break;
	  case _MESS:  PrintSFloat( Data.W,  256, 2 ); break;
    case _MESS10:PrintSFloat( Data.W,   10, 1 ); break;
	  case _MESS16:PrintSFloat( Data.W,   16, 1 ); break;
	  case _UMESS: PrintSFloat( Data.UW, 256, 2 ); break;
    case _STATUSW: PrintHex( Data.UW, 4 ); break;
    case _STATUSB: PrintHex( Data.UB, 2 ); break;

	  case _TIME:
      pBuf = toHexByte( highByte( Data.UW ), 2 );
      FlexPrint( pBuf );  
      pBuf = toHexByte( lowByte( Data.UW) , 2 );
      *(--pBuf) = ':';
      FlexPrint( pBuf );  
		  break;

	  case _TAG:
		  if( (Data.UB < 8U) && (Data.UB > 0U))
		    FlexPrint( (char *)Wochentag[ Data.UB - 1 ] );
		  break;

    case _BIG:
      pBuf  = OutBuf + OUTBUFLEN;
      *(--pBuf) = 0;
      pBuf = UWtoASCII( pBuf, ((uint16_t *)Data.P)[0], 3 );
      *(--pBuf) = Dezimalpunkt;
      pBuf = UWtoASCII( pBuf, ((uint16_t *)Data.P)[1], 1 );
      FlexPrint( pBuf );
      break;

	  case _VBIG:
      pBuf  = OutBuf + OUTBUFLEN;
      *(--pBuf) = 0;
      pBuf = UWtoASCII( pBuf, ((uint16_t *)Data.P)[0], 3 );
      *(--pBuf) = Dezimalpunkt;
      if( ((uint16_t *)Data.P)[2] == 0 )
        pBuf = UWtoASCII( pBuf, ((uint16_t *)Data.P)[1], 1 );
      else
        {
        pBuf = UWtoASCII( pBuf, ((uint16_t *)Data.P)[1], 3 );
        pBuf = UWtoASCII( pBuf, ((uint16_t *)Data.P)[2], 1 );
        }
      FlexPrint( pBuf );
		  break;

	  case _STATUS:
		  PrintHex( Data.UB, 2 );
      if(  (OutPut & OUT_SERIAL) && ((EEPromData.Mode == MODE_HEX) || (EEPromData.Mode == MODE_HEX_INFO) ) )
        {
		    FlexPrint( (char *)" Brenner: " );
        strcpy_P( OutBuf, (char*)pgm_read_word(&(StatusText[ Data.S.Kessel ]))); 
        FlexPrint( OutBuf );
		    FlexPrint( (char *)", Solarpumpe: " );
		    FlexPrint( (char *)( Data.S.SKPumpe ? "An": "Aus" ) );
        }
		  break;
   
  	default: if( OutPut & OUT_SERIAL )
  	    COMMAND_PORT.print( F("Error wrong type") );
	  }
  }


/*--------------------------------------------------------------------------
* PrintText - Gibt Text und Parameter aus
* Parameter: Nummer des Messwertes
* Ergebnis:  keins
*--------------------------------------------------------------------------*/

void PrintText( uint8_t Mess )

  {
  FlexPrint( (char *) " " );
  FlexPrint( (char *)HeizInfo[ Mess ].Text );
  FlexPrint( (char *)": " );
  PrintData( HeizInfo[ Mess ].Data, HeizInfo[ Mess ].Type );
  }


/*--------------------------------------------------------------------------
* PrintSelData - Gibt ausgewaehlte Parameter aus
* Parameter: Mask - 1 = Seriell, 2 = SD - Karte
*            Sep - Trennzeichen
* Ergebnis:  keins
*--------------------------------------------------------------------------*/

void PrintSelData( uint8_t Mask, char Sep )

  {
  uint8_t i;

  for( i=0; i<NumPara; i++ )
    if( EEPromData.Enable[ i ] & Mask )
      {
      PrintData( HeizInfo[ i ].Data, HeizInfo[ i ].Type );
      OutBuf[0] = Sep;
      OutBuf[1] = 0;
      FlexPrint( OutBuf );
      }

  PrintNewLine();
  }


#ifdef SD_LOGGING
/*--------------------------------------------------------------------------
* SaveSelData - Speichert ausgewählte Parameter in Logdatei
* Parameter: keine
* Ergebnis:  keins
*--------------------------------------------------------------------------*/

void SaveSelData( void )

  {
  OutPut = OUT_SD;

  dataFile = SD.open("datalog.txt", FILE_WRITE);
  if (dataFile)
    {
    PrintSelData( 2, SEPARATOR_SDC );
    dataFile.close();
    }

  OutPut = OUT_SERIAL;
  }


/*--------------------------------------------------------------------------
* SaveSelData - Log Message auf SD-Karte schreiben
* Parameter: keine
* Ergebnis:  keins
*--------------------------------------------------------------------------*/

void DataLogMsg( void )

  {
  uint8_t i;

  Dezimalpunkt = ',';
  dataFile = SD.open("datalog.txt", FILE_WRITE);
  if (dataFile)
    {
    dataFile.println( F("*** Logging Started ***") );
    for( i=0; i<NumPara; i++ )
      if( EEPromData.Enable[ i ] & 0x02 )
        {
        dataFile.write( HeizInfo[ i ].Text );
        dataFile.write( SEPARATOR_SDC );
        }

    dataFile.println();
    dataFile.close();
    }
  }
#endif /* #ifdef SD_LOGGING */


/*--------------------------------------------------------------------------
* PrintAllData - Gibt ausgew�hlte Parameter aus
* Parameter: keine
* Ergebnis:  keins
*--------------------------------------------------------------------------*/

void PrintAllData( void )

  {
  uint8_t i;


  Dezimalpunkt = ',';
  PrintNewLine();
  for( i=0; i<NumPara; i++ )
    {
    OutBuf[0] = (EEPromData.Enable[ i ] & 0x07) + '0';
    OutBuf[1] = 0;
    FlexPrint( &OutBuf[0] );
    FlexPrint( (char *)" - " );
    FlexPrint( (char *)HeizInfo[ i ].Text );
    FlexPrint( (char *)": " );
    PrintData( HeizInfo[ i ].Data, HeizInfo[ i ].Type );
    PrintNewLine();
    }

  PrintNewLine();
  }


/*--------------------------------------------------------------------------
* GetHex - 16Bit Hexadezimal Wert einlesen
* Parameter: pBuf - Zeiger auf String
*            pVal - Zeiger auf Ergebnis
* Ergebnis:  pBuf - neuer Zeiger auf String
*--------------------------------------------------------------------------*/

char *GetHex( char *pBuf, uint16_t *pVal )

  {
  uint8_t cnt = 4;
  *pVal = 0;

  while( *pBuf == ' ' )
    pBuf++;  

  do
    {
    if( (*pBuf >= '0') && (*pBuf <= '9') )
      *pVal = (*pVal << 4) + *(pBuf++) - '0';
    else if( (*pBuf >= 'A') && (*pBuf <= 'F') )
      *pVal = (*pVal << 4) + *(pBuf++) - 'A' + 10;
    else if( (*pBuf >= 'a') && (*pBuf <= 'f') )
      *pVal = (*pVal << 4) + *(pBuf++) - 'a' + 10;
    else
      break;
    }
  while( cnt-- > 0 );
  return( pBuf );
  }


/*--------------------------------------------------------------------------
* GetDec - 16Bit Dezimal Wert einlesen
* Parameter: pBuf - Zeiger auf String
*            pVal - Zeiger auf Ergebnis
* Ergebnis:  pBuf - neuer Zeiger auf String
*--------------------------------------------------------------------------*/

char *GetDec( char *pBuf, int16_t *pVal )

  {
  bool sig = false;
  uint8_t cnt = 5;
  *pVal = 0;

  while( *pBuf == ' ' )
    pBuf++;

  if( *pBuf == '-' )
    {
    sig = true;
    pBuf++;
    }

  do
    {
    if( (*pBuf >= '0') && (*pBuf <= '9') )
      *pVal = *pVal * 10 + *(pBuf++) - '0';
    else
      break;
    }
  while( cnt-- > 0 );

  if( sig )
    *pVal = -(*pVal);
  return( pBuf );
  }

